#!/usr/bin/env python3
import sys
import os
import pathlib
import argparse
from inspect import cleandoc


hook_text = cleandoc(
    f"""
    {{
        "version": "1.0.0",
        "hook": {{
            "path": "{os.path.join(sys.prefix,'bin','hook_tool')}",
            "args": ["hook_tool"]
        }},
        "when": {{
            "annotations": {{
               "podman_hpc.hook_tool": "true"
            }}
        }},
        "stages": ["prestart"]
    }}

    """
)


def write_hook(hooksd, hook, filemode):
    hook_path = os.path.join(hooksd, hook)
    os.makedirs(hooksd, exist_ok=True)
    try:
        with open(hook_path, filemode) as fid:
            fid.write(hook_text)
        print(f"Successfully wrote hook configuration at {hook_path}")
    except FileExistsError:
        print(f"Hook is already configured at {hook_path}. No changes made.")


def main():
    p = argparse.ArgumentParser(
        prog="configure_hooks",
        description="Write a hooks.json and install to hooksd for podman-hpc.",
    )
    p.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force overwrite of existing hook file.",
    )
    p.add_argument(
        "--hooksd",
        type=pathlib.Path,
        default=os.path.join(
            sys.prefix, "share", "containers", "oci", "hooks.d"
        ),
        help="Path to write hook file.",
    )

    ns = p.parse_args()
    hook = "02-hook_tool.json"
    filemode = "w" if ns.force else "x"

    write_hook(ns.hooksd, hook, filemode)


if __name__ == "__main__":
    main()



import os
import re
import click
import warnings
import inspect
import typing as t
import functools

try:
    from argparse import ArgumentParser as _

    _ = _(exit_on_error=True)
    import argparse
except TypeError:
    from . import argparse_exit_on_error as argparse


class DefaultCommandGroup(click.Group):
    def __init__(self, *args, hide_default=False, **attrs):
        super().__init__(*args, **attrs)
        self.default_command_fn = None
        self.hide_default = hide_default

    def set_default_command(self, cmd, name=None):
        if not self.default_command_fn:
            name = name or cmd.name
            if name is None:
                raise TypeError("Command has no name.")
            self.default_command_fn = cmd
        else:
            warnings.warn(
                f"{self.name}.default_command_fn is already set: " +
                f"{self.default_command_fn}",
                UserWarning,
                stacklevel=2,
            )

    def add_command(self, cmd, name=None, default=False):
        super().add_command(cmd, name)
        if default:
            self.set_default_command(cmd)

    def default_command(self, *args, **kwargs):
        from click.decorators import command

        if self.command_class and kwargs.get("cls") is None:
            kwargs["cls"] = self.command_class

        func: t.Optional[t.Callable] = None

        if args and callable(args[0]):
            assert (
                len(args) == 1 and not kwargs
            ), "Use 'command(**kwargs)(callable)' to provide arguments."
            (func,) = args
            args = ()

        def decorator(f: t.Callable[..., t.Any]) -> click.Command:
            cmd: click.Command = command(*args, **kwargs)(f)
            self.set_default_command(cmd)
            self.add_command(cmd)
            return cmd

        if func is not None:
            return decorator(func)

        return decorator

    def get_command(
        self, ctx: click.Context, cmd_name: str
    ) -> t.Optional[click.Command]:
        return self.commands.get(cmd_name, self.default_command_fn)

    def list_commands(self, ctx: click.Context) -> t.List[str]:
        if self.default_command_fn and self.hide_default:
            return [
                cmd
                for cmd in sorted(self.commands)
                if not (
                    cmd == self.default_command_fn.name and self.hide_default
                )
            ]
        else:
            return sorted(self.commands)


def customize_help(cls):
    """Decorator to add custom help formatting on _click classes.

    This decorator can be placed before class declarations for classes
    inheriting from click.Command.

    The class __init__ method is modified to take an additional keyword
    parameter, `custom_format`.  The value should be a function which
    accepts three parameters (self, ctx, formatter), following the
    convention and syntax of click.Command.format_help().  The decorator
    additionally adds other convenience methods to extract description
    and example text from the docstring.

    If no custom formatting function is provided when the decorated
    class is instantiated, it may be provided later by setting the
    custom_format attribute on the instantiated object.

    If no custom_format is set when the instance attempts to display
    the help message, then the object will attempt to inherit a format
    from parents in it's click calling context, until either (1) a
    custom_format is found, (2) an undecorated parent is encountered
    while ascending the context tree, or (3) the root of the calling
    context is reached.  If no custom formatting function is present
    or inherited, then the decorated class will revert to standard
    click help formatting.

    """
    # wrappers to modify, otherwise just replace
    def mod__init__(cls_init):
        @functools.wraps(cls_init)
        def wrapper(self, *args, custom_format=None, **kwargs):
            self.custom_format = custom_format
            return cls_init(self, *args, **kwargs)

        return wrapper

    def get_custom_format(self, ctx):
        if self.custom_format is not None:
            return self.custom_format
        elif ctx.parent is not None:
            return getattr(
                ctx.parent.command, "get_custom_format", lambda x: None
            )(ctx.parent)
        else:
            return None

    def format_short_help(self, ctx, formatter):
        if getattr(self, "short_help", None) is None:
            self.short_help = self.help.splitlines()[0] if self.help else None
        if getattr(self, "short_help", None):
            formatter.write_text(self.short_help)

    def format_description(self, ctx, formatter):
        if getattr(self, "description", None) is None:
            self.description = (
                inspect.cleandoc(re.sub("^.+", "", self.help))
                if self.help
                else None
            )
        if getattr(self, "description", None):
            with formatter.section("Description"):
                formatter.write_text(self.description)

    def mod_format_help(cls_format_help):
        @functools.wraps(cls_format_help)
        def wrapper(self, ctx, formatter):
            custom_format_help = self.get_custom_format(ctx)
            if custom_format_help is not None:
                custom_format_help(self, ctx, formatter)
            else:
                cls_format_help(self, ctx, formatter)

        return wrapper

    cls.__init__ = mod__init__(cls.__init__)
    cls.get_custom_format = get_custom_format
    cls.format_short_help = format_short_help
    cls.format_description = format_description
    cls.format_help = mod_format_help(cls.format_help)
    return cls


@customize_help
class TemplateHelpCommand(click.Command):
    pass


@customize_help
class PassthroughGroup(DefaultCommandGroup):
    def __init__(self, *args, passthrough=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.command_class = TemplateHelpCommand
        if passthrough is not None:
            self.init_passthrough(passthrough)

    def init_passthrough(self, passthrough_bin):
        """Generate a default command for this group,
        which calls the specified passthrough bin."""
        pass


def parametersFromPassthrough(cmd, passthrough):
    pass


def filterValidOptions(options, subcmd, option_regex=None):
    """Filter invalid arguments from an argument list
    for a given subcommand based on its --help text."""
    if not option_regex:
        option_regex = re.compile(r"^\s*(?:(-\w), )?(--\w[\w\-]+)(?:\s(\w+))?")

    # create a dummy parser and populate it with valid option flags
    p = argparse.ArgumentParser(exit_on_error=False, add_help=False)
    with os.popen(" ".join(subcmd)) as f:
        for line in f:
            opt = option_regex.match(line)
            if opt:
                action = "store" if opt.groups()[2] else "store_true"
                flags = [flag for flag in opt.groups()[:-1] if flag]
                p.add_argument(*flags, action=action)

    # remove unknown options from the given options
    valid_options = options.copy()
    unknowns = p.parse_known_args(valid_options)[1]
    # some args may match unknowns but actually be valid because they
    # provide a value for a valid flag, store these "safe" indices here
    uk_safe_index = {}
    # iterate until we've accounted for all the unknown options
    while unknowns:
        ukd = {}  # candidate indices of where to remove unknowns
        for uk in set(unknowns):
            ukd[uk] = [
                idx
                for idx, opt in enumerate(valid_options)
                if (opt == uk and idx not in uk_safe_index.get(uk, []))
            ]
        uk = unknowns.pop(0)
        # find and remove an invalid occurence of uk
        while True:
            valid_options_tmp = valid_options.copy()
            valid_options_tmp.pop(ukd[uk][0])
            try:
                if p.parse_known_args(valid_options_tmp)[1] == unknowns:
                    valid_options.pop(ukd[uk][0])
                    break
            except argparse.ArgumentError:
                pass
            uk_safe_index.setdefault(uk, []).append(ukd[uk].pop(0))
    return valid_options



import sys
import os
import shutil
import toml
import re
from yaml import load
from yaml import FullLoader
from copy import deepcopy
from glob import glob

_ENV_PREFIX = "PODMANHPC"
_MOD_ENV = f"{_ENV_PREFIX}_MODULES_DIR"
_HOOKS_ANNO = "podman_hpc.hook_tool"
_CONF_ENV = f"{_ENV_PREFIX}_CONFIG_FILE"


class SiteConfig:
    """
    Class to represent site specific configurations for Podman-HPC.

    Precedence order lowest to highest:
    - Built-in defaults
    - Config file
    - Environment variables
    """

    _default_conf_file = "/etc/podman_hpc/podman_hpc.yaml"
    _valid_params = ["podman_bin", "mount_program", "modules_dir",
                     "shared_run_exec_args", "shared_run_command",
                     "graph_root", "run_root",
                     "additional_stores", "hooks_dir",
                     "localid_var", "tasks_per_node_var", "ntasks_pattern",
                     "config_home", "mksquashfs_bin",
                     "wait_timeout", "wait_poll_interval",
                     "use_default_args",
                     ]
    _valid_templates = ["shared_run_args_template",
                        "graph_root_template",
                        "run_root_template",
                        "additional_stores_template",
                        "config_home_template"
                        ]
    _uid = os.getuid()
    _xdg_base = f"/tmp/{_uid}_hpc"
    config_home = f"{_xdg_base}/config"
    run_root = _xdg_base
    additional_stores = []
    hooks_dir = f"{sys.prefix}/share/containers/oci/hooks.d"
    graph_root = f"{_xdg_base}/storage"
    squash_dir = os.environ.get(
        "SQUASH_DIR", f'{os.environ.get("SCRATCH", "/tmp")}/storage'
    )
    modules_dir = "/etc/podman_hpc/modules.d"
    shared_run_exec_args = ["-e", "SLURM_*", "-e", "PALS_*", "-e", "PMI_*"]
    use_default_args = True
    shared_run_command = ["sleep", "infinity"]
    podman_bin = "podman"
    mount_program = "fuse-overlayfs-wrap"
    runtime = "runc"
    localid_var = "SLURM_LOCALID"
    tasks_per_node_var = "SLURM_STEP_TASKS_PER_NODE"
    ntasks_pattern = r'[0-9]+'
    mksquashfs_bin = "mksquashfs.static"
    wait_poll_interval = 0.2
    wait_timeout = 10
    shared_run = False
    source = dict()

    def __init__(self, squash_dir=None, log_level=None):

        # getlogin may fail on a compute node
        try:
            self.user = os.getlogin()
        except Exception:
            self.user = os.environ["USER"]
        # TODO: move these as a test at the end
        self.podman_bin = self.trywhich("podman")
        self.mount_program = self.trywhich("fuse-overlayfs-wrap")
        try:
            self.runtime = self.trywhich("crun")
        except OSError:
            self.runtime = self.trywhich("runc")
        # self.options = []
        if squash_dir:
            self.squash_dir = squash_dir
        self.conf_file_data = {}
        self._read_config_file()
        for param in self._valid_templates:
            self._check_and_set(param)
        for param in self._valid_params:
            self._check_and_set(param)

        self.read_site_modules()

        if isinstance(self.wait_poll_interval, str):
            self.wait_poll_interval = \
                float(self.wait_poll_interval)
        if isinstance(self.wait_timeout, str):
            self.wait_timeout = float(self.wait_timeout)

        if self.use_default_args is True:
            self.default_args = [
                    "--root", self.graph_root,
                    "--runroot", self.run_root,
                    "--storage-opt",
                    f"mount_program={self.mount_program}",
                    "--cgroup-manager", "cgroupfs",
                    ]
            self.default_run_args = [
                    "--storage-opt",
                    "ignore_chown_errors=true",                    
                    "--storage-opt",
                    f"additionalimagestore={self.additionalimagestore()}",
                    "--hooks-dir", self.hooks_dir,
                    "--env", f"{_MOD_ENV}={self.modules_dir}",
                    "--annotation", f"{_HOOKS_ANNO}=true",
                    "--security-opt", "seccomp=unconfined",
                    ]
            self.default_build_args = [
                    "--hooks-dir", self.hooks_dir,
                    "--env", f"{_MOD_ENV}={self.modules_dir}",
                    "--annotation", f"{_HOOKS_ANNO}=true",
                    ]
            self.default_pull_args = [
                    "--storage-opt",
                    "ignore_chown_errors=true",
                    ]
            self.default_images_args = [
                    "--storage-opt",
                    f"additionalimagestore={self.additionalimagestore()}",
                    ]
        else:
            self.default_args = []
            self.default_run_args = []
            self.default_build_args = []
            self.default_pull_args = []
            self.default_images_args = []
        
        self.log_level = log_level

    def dump_config(self):
        """
        Debug method to dump the configuration
        """

        for param in self._valid_params + ["active_modules"]:
            val = getattr(self, param)
            source = ""
            if param in self.source:
                source = f" ({self.source[param]})"
            # source = self.source.get(param, "unknown")
            if isinstance(val, list):
                print(f"{param} {source}: (list)")
                for item in val:
                    print(f" - {item}")
            else:
                print(f"{param}{source}: {val}")

    @staticmethod
    def trywhich(cmd, *args, **kwargs):
        """
        Use which and raise an error if it can't be found.
        """
        res = shutil.which(cmd, *args, **kwargs)
        if res is None:
            raise OSError(
                f"No {cmd} binary found in {kwargs.get('path','PATH')}"
            )
        return res

    def _check_and_set(self, attr: str, envname=None, parname=None):
        """
        Helper function to apply the right precedence
        """
        source = "default"
        if not parname:
            parname = attr
        if not envname:
            uppar = attr.upper()
            envname = f"{_ENV_PREFIX}_{uppar}"

        setval = False
        if envname in os.environ:
            setval = True
            newval = os.environ[envname]
            source = f"env: {envname}"
        elif parname in self.conf_file_data:
            setval = True
            newval = self.conf_file_data[parname]
            source = f"file: {parname}"
        if setval and attr.endswith("_template"):
            setattr(self, attr, newval)
            newval = self._apply_template(newval)
            attr = attr.replace("_template", "")

        if setval:
            # Expand to a list if the type should be a list
            # Assumes a common seperated string
            if isinstance(getattr(self, attr), list) and \
               isinstance(newval, str):
                newval = newval.split(',')
            setattr(self, attr, newval)
            self.source[attr] = source
        else:
            if attr not in self.source:
                self.source[attr] = source

    def _apply_template(self, templ):
        """
        Helper function to convert templates
        """
        def _templ(val):
            val = val.replace("{{ user }}", self.user)
            val = val.replace("{{ uid }}", str(self._uid))
            for pat in re.findall(r'{{ env\.[A-Za-z0-9]+ }}', val):
                envname = pat.replace("{{ env.", "").replace(" }}", "")
                val = val.replace(pat, os.environ[envname])
            return val

        if isinstance(templ, list):
            newlist = []
            for item in templ:
                newlist.append(_templ(item))
            return newlist
        elif isinstance(templ, str):
            return _templ(templ)
        else:
            raise ValueError("Can handle template type")

    def _read_config_file(self):
        """
        Read the global config file
        """
        config_file = os.environ.get(_CONF_ENV, self._default_conf_file)
        if not os.path.exists(config_file):
            return
        self.conf_file_data = load(open(config_file), Loader=FullLoader)
        for p in self.conf_file_data:
            if p not in self._valid_params and p not in self._valid_templates:
                raise ValueError(f"Unrecongnized Option: {p}")

    def get_default_store_conf(self):
        """
        Generate a default storage conf object
        """
        return {
            "storage": {
                "driver": "overlay",
                "graphroot": self.graph_root,
                "runroot": self.run_root,
                "options": {
                    "size": "",
                    "remap-uids": "",
                    "remap-gids": "",
                    "ignore_chown_errors": "true",
                    "remap-user": "",
                    "remap-group": "",
                    "mount_program": self.mount_program,
                    "mountopt": "",
                    "overlay": {},
                },
            }
        }

    def get_default_containers_conf(self):
        """
        Generate a default containers conf object
        """
        return {
            "engine": {
                "cgroup_manager": "cgroupfs",
            },
            "containers": {
                "seccomp_profile": "unconfined",
            },
        }

    def additionalimagestore(self):
        ais = [self.squash_dir]
        ais.extend(self.additional_stores)
        return ','.join(ais)

    def config_storage(self, additional_stores=None):
        """
        Create a storage conf object
        """
        sc = self.get_default_store_conf()
        ais = sc["storage"]["options"].setdefault("additionalimagestores", [])
        ais.append(self.squash_dir)
        if additional_stores:
            ais.extend(additional_stores.split(","))
        self.storage_conf = sc

    def config_containers(self):
        """
        Create a container conf object
        """
        cc = self.get_default_containers_conf()
        self.container_conf = cc

    def config_env(self, hpc):
        """
        Generate the environment setup

        TODO: Redo this
        """
        new_env = deepcopy(os.environ)
        if hpc:
            new_env["XDG_CONFIG_HOME"] = self.config_home
            new_env.pop("XDG_RUNTIME_DIR", None)
        self.env = new_env

    def _write_conf(self, filename, data, overwrite=False):
        """
        Write out a conf file
        """
        os.makedirs(
            f"/tmp/containers-user-{self._uid}/containers", exist_ok=True
        )
        os.makedirs(f"{self.config_home}/containers", exist_ok=True)
        fp = os.path.join(self.config_home, "containers", filename)
        if not os.path.exists(fp) or overwrite:
            with open(fp, "w") as f:
                toml.dump(data, f)

    def export_storage_conf(self, filename="storage.conf", overwrite=False):
        """
        Write out the storage.conf file
        """
        self._write_conf(filename, self.storage_conf, overwrite=overwrite)

    def export_containers_conf(
        self, filename="containers.conf", overwrite=False
    ):
        """
        Write out the containers.conf file
        """
        self._write_conf(filename, self.container_conf, overwrite=overwrite)

    def get_cmd_extensions(self, subcommand, args):
        """
        Generate the podman command-line parameters

        subcomand: run, images, etc
        """
        cmds = []
        cmds.extend(self.default_args)
        if subcommand == "run":
            cmds.extend(self.default_run_args)
        elif subcommand == "build":
            cmds.extend(self.default_build_args)
        elif subcommand == "pull":
            cmds.extend(self.default_pull_args)
        elif subcommand == "images":
            cmds.extend(self.default_images_args)
        else:
            pass
        for mod, mconf in self.sitemods.get(subcommand, {}).items():
            if 'cli_arg' not in mconf:
                continue
            cli_arg = mconf["cli_arg"].replace("-", "_")
            if args.get(cli_arg, False):
                cmds.extend(mconf.get("additional_args", []))
                cmds.extend(["-e", f"{mconf['env']}=1"])
                if mconf.get("shared_run"):
                    self.shared_run = True
        if self.log_level:
            cmds.extend(["--log-level", self.log_level])
        return cmds

    # TODO - This manually indicates that site modules apply to podman
    # subcommand `run`, however we should bake this into yamls themselves
    # and then edit podman_hpc.siteconfig.SiteConfig.read_site_modules()
    # to parse appropriately.  This would allow adding site-specific default
    # flags for any podman subcommand.
    def read_site_modules(self):
        mods = {}
        active_mods = []
        for modfile in glob(f"{self.modules_dir}/*.yaml"):
            mod = load(open(modfile), Loader=FullLoader)
            mods[mod["name"]] = mod
            active_mods.append(modfile)
        self.active_modules = active_mods
        self.sitemods = {"run": mods, "shared-run": mods}



import podman_hpc.podman_hpc as phpc
import sys
import os
import pytest
import json


@pytest.fixture
def fix_paths(monkeypatch):
    test_dir = os.path.dirname(__file__)
    mock_dir = os.path.join(test_dir, "mock_bin")
    bin_dir = os.path.join(test_dir, "..", "bin")
    monkeypatch.setenv("PATH", mock_dir, prepend=os.pathsep)
    monkeypatch.setenv("PATH", bin_dir, prepend=os.pathsep)


@pytest.fixture
def mock_exit(monkeypatch):
    def mock_exit_func(*args):
        return 0
    monkeypatch.setattr(sys, "exit", mock_exit_func)


@pytest.fixture
def mock_podman(monkeypatch, tmp_path):
    test_dir = os.path.dirname(__file__)
    mock_pod = os.path.join(test_dir, "mock_bin", "mock_podman")
    monkeypatch.setenv("PODMANHPC_PODMAN_BIN", mock_pod)
    mock_out = os.path.join(tmp_path, "mock.out")
    monkeypatch.setenv("MOCK_OUT", mock_out)
    return mock_out


def test_help(monkeypatch, fix_paths, capsys, mock_exit):
    sys.argv = ["podman_hpc", "--help"]
    phpc.main()
    captured = capsys.readouterr()
    assert "podman-hpc" in captured.out
    assert "Podman help page follows" in captured.out


def test_run_help(monkeypatch, fix_paths, capsys, mock_exit):
    sys.argv = ["podman_hpc", "run", "--help"]
    phpc.main()
    captured = capsys.readouterr()
    assert "podman-hpc" in captured.out
    assert "Podman options" in captured.out


def test_run(monkeypatch, fix_paths, mock_exit):
    sys.argv = ["podman_hpc", "run", "-it", "--rm", "ubuntu"]
    uid = os.getuid()
    global args_passed
    args_passed = []

    def mock_execve(bin, args, path):
        global args_passed
        args_passed.append(args)
        return 0

    monkeypatch.setattr(os, "execve", mock_execve)
    phpc.main()
    # TODO: Add more checks
    last = args_passed[-1]
    assert "run" in last
    assert f"/tmp/{uid}_hpc/storage" in last
    assert f"/tmp/{uid}_hpc" in last
    assert "cgroupfs" in last
    assert "--hooks-dir" in last
    assert "PODMANHPC_MODULES_DIR=/etc/podman_hpc/modules.d" in last
    assert "fatal" in last
    assert "podman_hpc.hook_tool=true"
    assert sys.argv[2:] == last[-3:]
    assert "seccomp=unconfined"


def test_shared_run(monkeypatch, fix_paths, mock_podman, mock_exit):
    sys.argv = ["podman_hpc", "shared-run", "-it", "--rm",
                "--volume", "/a:/b", "ubuntu", "uptime"]
    monkeypatch.setenv("SLURM_LOCALID", "0")
    monkeypatch.setenv("SLURM_STEP_TASKS_PER_NODE", "1")
    phpc.main()
    run = None
    with open(mock_podman) as f:
        for line in f.read().split("\n"):
            items = line.split(" ")
            if items[0] == "run":
                run = items
            elif "exec --root" in line:
                exec = items
    uid = os.getuid()
    assert run is not None
    assert "--root" in run
    assert f"/tmp/{uid}_hpc/storage" in run
    assert "--name" in run
    assert "ubuntu" in run
    assert exec is not None
    assert exec[-1] == "uptime"
    assert "SLURM_*" in exec
    assert "PALS_*" in exec
    assert "--volume" not in exec
    assert "--rm" not in exec


def test_run_fail(monkeypatch, fix_paths, mock_podman, mock_exit):
    sys.argv = ["podman_hpc", "shared-run", "-it", "--rm",
                "-e", "FAILME=1", "ubuntu", "uptime"]
    monkeypatch.setenv("SLURM_LOCALID", "0")
    monkeypatch.setenv("SLURM_STEP_TASKS_PER_NODE", "1")
    monkeypatch.setenv("PODMANHPC_WAIT_TIMEOUT", "0.5")
    monkeypatch.setenv("MOCK_FAILURE", "1")
    phpc.main()
    run = None
    out = open(mock_podman).read()
    assert "run --rm" in out
    assert "exec --root" not in out
    fn = f"/tmp/uid-{os.getuid()}-pid-{os.getppid()}.txt"


def test_shared_run_auto(monkeypatch, fix_paths, mock_podman, mock_exit):
    sys.argv = ["podman_hpc", "run", "-it", "--rm", "--mpi",
                "--volume", "/a:/b", "ubuntu", "uptime"]
    monkeypatch.setenv("SLURM_LOCALID", "0")
    monkeypatch.setenv("SLURM_STEP_TASKS_PER_NODE", "1")
    test_dir = os.path.dirname(__file__)
    modules_dir = os.path.join(test_dir, "..", "etc", "modules.d")
    monkeypatch.setenv("PODMANHPC_MODULES_DIR", modules_dir)
    phpc.main()
    run = None
    with open(mock_podman) as f:
        for line in f:
            items = line.split()
            if items[0] == "run":
                run = items
            elif items[0] == "exec":
                exec = items
    uid = os.getuid()
    assert run is not None
    assert "ubuntu" in run
    assert exec is not None
    assert exec[-1] == "uptime"
    assert "SLURM_*" in exec
    assert "PALS_*" in exec


def test_pull(monkeypatch, fix_paths, mock_podman, mock_exit,
              tmp_path, capsys):
    sys.argv = ["podman_hpc", "--squash-dir", str(tmp_path),
                "pull", "alpine"]
    tdir = os.path.dirname(__file__)
    src = os.path.join(tdir, "storage")
    monkeypatch.setenv("PODMANHPC_GRAPH_ROOT", src)
    phpc.main()
    captured = capsys.readouterr()
    assert "Migrating" in captured.out
    assert str(tmp_path) in captured.out
    out = open(mock_podman).read()
    assert "pull " in out
    assert src in out
    assert captured.err == ""
    imagej = os.path.join(tmp_path, "overlay-images",
                          "images.json")
    d = json.load(open(imagej))
    id = "9c6f0724472873bb50a2ae67a9e7adcb57673a183cea8b06eb778dca859181b5"
    assert d[0]['id'] == id
    run = out.split("\n")[-2].split()
    assert "/mksq" in run
    assert "/sqout" in run
    sys.argv = ["podman_hpc", "--squash-dir", str(tmp_path),
                "rmsqi", "alpine"]
    phpc.main()
    d = json.load(open(imagej))
    assert len(d) == 0


def test_migrate(fix_paths, mock_podman, mock_exit, tmp_path):
    sys.argv = ["podman_hpc", "--squash-dir",
                str(tmp_path), "migrate", "ubuntu"]
    phpc.main()
    assert os.path.exists(os.path.join(tmp_path, "overlay"))
    # TODO: Add more checks


def test_modules(monkeypatch, fix_paths, mock_exit):
    sys.argv = ["podman_hpc", "run", "-it", "--rm", "--gpu",
                "ubuntu"]
    global args_passed
    args_passed = []

    def mock_execve(bin, args, path):
        global args_passed
        args_passed.append(args)
        return 0

    monkeypatch.setattr(os, "execve", mock_execve)
    test_dir = os.path.dirname(__file__)
    modules_dir = os.path.join(test_dir, "..", "etc", "modules.d")
    monkeypatch.setenv("PODMANHPC_MODULES_DIR", modules_dir)
    phpc.main()
    assert "--gpu" not in args_passed[0]
    assert "ENABLE_GPU=1" in args_passed[0]



import podman_hpc.hook_tool as ht
import podman_hpc.configure_hooks as ch
import sys
import io
import os
import json
import ctypes
import time
import pytest


_libc = ctypes.CDLL(ctypes.util.find_library('c'), use_errno=True)


@pytest.mark.skipif(not os.path.exists("/proc"),
                    reason="requires running on linux system")
def test_hook(monkeypatch, tmp_path):
    tdir = os.path.dirname(__file__)
    # conf = os.path.join(tdir, "test.yaml")
    conf = os.path.join(tdir, "modules.d")
    log_file = os.path.join(tmp_path, "log.out")
    rdir = os.path.join(tmp_path, "root")
    os.mkdir(rdir)
    os.mkdir(os.path.join(rdir, "etc"))
    env = ["ENABLE_MODULE1=1",
           f"LOG_PLUGIN={log_file}",
           f"{ht._MOD_ENV}={conf}",
           ]
    cconf = {
        "root": {
            "path": rdir
            },
        "process": {"env": env}
    }
    with open(os.path.join(tmp_path, "config.json"), "w") as f:
        json.dump(cconf, f)
    pid = os.fork()
    if pid == 0:
        uidmapfile = '/proc/self/uid_map'
        uidmap = "0 %d 1" % os.getuid()
        _libc.unshare(0x00020000 | 0x10000000 | 0x20000000)
        print("Writing uidmap = '%s' to '%s'" % (uidmap, uidmapfile))
        with open(uidmapfile, 'w') as file:
            file.write(uidmap)
        time.sleep(4)
        os._exit(0)
    time.sleep(1)
    ht.setns(pid, "user")

    hconf = {
            'pid': pid,
            'annotations': {}
           }
    sys.stdin = io.StringIO(json.dumps(hconf))
    here = os.getcwd()
    sys.argv = [sys.argv[0]]
    os.chdir(tmp_path)

    def mock_chroot(dir):
        return 0

    monkeypatch.setattr(os, "chroot", mock_chroot)
    ht.main()
    os.chdir(here)
    motd = os.path.join(rdir, "a/motd")
    assert os.path.exists(motd)
    null = os.path.join(rdir, "null")
    assert os.path.exists(null)
    hosts = os.path.join(rdir, "etc/hosts")
    assert os.path.exists(hosts)
    null2 = os.path.join(rdir, "null2")
    assert not os.path.exists(null2)
    os.wait()
    _libc.umount2(null, 2)
    assert os.path.exists(log_file)


def test_conf_hook(monkeypatch, tmp_path, capsys):
    hook_out = os.path.join(tmp_path, "02-hook_tool.json")

    def mock_exit(exit_code):
        return 0

    sys.argv = ["podman_hpc", "--hooksd", str(tmp_path)]
    monkeypatch.setattr(sys, "exit", mock_exit)
    ch.main()
    captured = capsys.readouterr()
    assert "Successfully" in captured.out
    ho = json.load(open(hook_out))
    assert "version" in ho



from argparse import *
from argparse import ArgumentParser as _ArgumentParser


class ArgumentParser(_ArgumentParser):
    def __init__(self, *args, exit_on_error=True, **kwargs):
        self._exit_on_error = exit_on_error
        super().__init__(*args, **kwargs)

    def error(self, *args, **kwargs):
        if self._exit_on_error:
            super().error(*args, **kwargs)



import podman_hpc.siteconfig as config
import os
import pytest


@pytest.fixture
def fix_paths(monkeypatch):
    test_dir = os.path.dirname(__file__)
    mock_dir = os.path.join(test_dir, "mock_bin")
    bin_dir = os.path.join(test_dir, "..", "bin")
    monkeypatch.setenv("PATH", mock_dir, prepend=os.pathsep)
    monkeypatch.setenv("PATH", bin_dir, prepend=os.pathsep)


@pytest.fixture
def conf_file(monkeypatch):
    test_dir = os.path.dirname(__file__)
    test_conf = os.path.join(test_dir, "test_conf.yaml")
    monkeypatch.setenv(config._CONF_ENV, test_conf)
    monkeypatch.setenv("FOO", "BAR")


def test_conf_defaults(fix_paths):
    conf = config.SiteConfig(squash_dir="/tmp")
    uid = os.getuid()
    assert conf.run_root == f"/tmp/{uid}_hpc"
    assert conf._xdg_base == f"/tmp/{uid}_hpc"
    assert conf.config_home == f"/tmp/{uid}_hpc/config"
    assert conf.get_default_store_conf() is not None
    assert conf.get_default_containers_conf() is not None
    assert f"/tmp/{uid}_hpc/storage" in conf.default_args
    conf.config_env(True)
    assert conf.env is not None


def test_conf_file(fix_paths, conf_file, monkeypatch):
    conf = config.SiteConfig(squash_dir="/tmp")
    uid = os.getuid()
    user = os.getlogin()
    assert conf.podman_bin == "/bin/sleep"
    assert conf.mount_program == "/bin/sleep"
    assert "SLURM_*" in conf.shared_run_exec_args
    assert "bogus" in conf.shared_run_command
    assert user in conf.graph_root
    assert str(uid) in conf.run_root
    assert f"/imagedir/{user}/storage" == conf.graph_root
    assert f"/tmp/{uid}/run" == conf.run_root
    # This test environment variable substition and
    # environment variables precedence
    monkeypatch.setenv("FOO", "BAR")
    tmpl = "/tmp/{{ env.FOO }}"
    monkeypatch.setenv("PODMANHPC_RUN_ROOT_TEMPLATE", tmpl)
    conf = config.SiteConfig(squash_dir="/tmp")
    assert "/tmp/BAR" == conf.run_root
    assert f"/images/{user}/storage" in conf.default_args
    assert f"/tmp/{uid}/" in conf.default_args


def test_conf_modules(fix_paths, monkeypatch):
    test_dir = os.path.dirname(__file__)
    modules_dir = os.path.join(test_dir, "..", "etc", "modules.d")
    monkeypatch.setenv("PODMANHPC_MODULES_DIR", modules_dir)
    conf = config.SiteConfig(squash_dir="/tmp", log_level="DEBUG")
    assert len(conf.active_modules) > 0
    args = conf.get_cmd_extensions("run", {"gpu": True})
    assert "ENABLE_GPU=1" in args


def test_conf_prec(conf_file, fix_paths, monkeypatch):
    test_dir = os.path.dirname(__file__)
    modules_dir = os.path.join(test_dir, "..", "etc", "modules.d")
    monkeypatch.setenv("PODMANHPC_MODULES_DIR", modules_dir)
    conf = config.SiteConfig(squash_dir="/tmp")
    assert conf.modules_dir == modules_dir


def test_conf_funcs(fix_paths, monkeypatch, tmp_path, capsys):
    monkeypatch.setenv("PODMANHPC_CONFIG_HOME", str(tmp_path))
    conf = config.SiteConfig(squash_dir="/tmp")
    conf.config_containers()
    conf.config_storage(additional_stores="/a:/b")
    conf.export_containers_conf()
    conf.export_storage_conf()
    sc = os.path.join(tmp_path, "containers", "storage.conf")
    assert os.path.exists(sc)
    sc = os.path.join(tmp_path, "containers", "containers.conf")
    assert os.path.exists(sc)
    conf.dump_config()
    captured = capsys.readouterr()
    assert "SLURM_LOCALID" in captured.out


def test_typing(fix_paths, monkeypatch, tmp_path):
    monkeypatch.setenv("PODMANHPC_CONFIG_HOME", str(tmp_path))
    monkeypatch.setenv("PODMANHPC_DEFAULT_ARGS", "a,b")
    monkeypatch.setenv("PODMANHPC_DEFAULT_RUN_ARGS_TEMPLATE",
                       "{{ uid }},{{ user }}")
    conf = config.SiteConfig(squash_dir="/tmp")
    assert isinstance(conf.default_args, list)
    assert conf.default_args == ["a", "b"]
    uid = os.getuid()
    user = os.getlogin()
    assert conf.default_run_args == [str(uid), user]



#!/usr/bin/env python

import sys
import os
import socket
import re
import time
import click
from . import click_passthrough as cpt
from .migrate2scratch import MigrateUtils
from .migrate2scratch import ImageStore
from .siteconfig import SiteConfig
from multiprocessing import Process
from subprocess import Popen, PIPE


def podman_devnull(cmd, conf):
    """
    Run a command and ignore the output.
    Returns the exit code
    """
    newcmd = [conf.podman_bin]
    newcmd.extend(conf.get_cmd_extensions(cmd[0], None))
    newcmd.extend(cmd)
    proc = Popen(newcmd, stdout=PIPE, stderr=PIPE)
    proc.communicate()
    return proc.returncode


def pmi_fd():
    """
    This method kicks in if PMI_FD is set.  If set,
    this will dup that file descriptor to fd 3,
    set PMI_FD to 3, and add a new cli arg to pass
    the file descriptor.

    Note: This is a bit of a one off and we should
    find a cleaner solution when there is time.
    """

    if "PMI_FD" not in os.environ:
        return []
    pmifd = int(os.environ['PMI_FD'])
    os.dup2(pmifd, 3)
    os.set_inheritable(3, True)
    os.environ['PMI_FD'] = "3"
    return ["--preserve-fds", "1"]


# function to specify help message formatting to mimic the podman help page.
# follows the style of click.Command.format_help()
# this will be inherited by subcommands created with @podhpc.command()
def podman_format(self, ctx, formatter):
    self.format_short_help(ctx, formatter)
    self.format_description(ctx, formatter)
    formatter.write_paragraph()
    self.format_usage(ctx, formatter)
    self.format_options(ctx, formatter)
    if self.epilog:
        with formatter.section("Podman help page follows"):
            self.format_epilog(ctx, formatter)


# parse the `podman --help` page so it can be added as a custom epilog to
# main command help
initenv = os.environ
if "XDG_RUNTIME_DIR" in os.environ:
    os.environ.pop("XDG_RUNTIME_DIR")
with os.popen("podman --help") as fid:
    try:
        text = re.sub(
            "^.+(?=(Available Commands))", "\b\n", fid.read(), flags=re.DOTALL
        )
        podman_epilog = re.sub(r"(\n\s*\n)(?=\S)", "\n\n\b\n", text)
    except Exception:
        podman_epilog = "For additional commands please see `podman --help`."
os.environ = initenv

# decorator so that subcommands can request to receive SiteConfig object
pass_siteconf = click.make_pass_decorator(SiteConfig, ensure=True)


# podman-hpc command #######################################################
@click.group(
    cls=cpt.PassthroughGroup,
    custom_format=podman_format,
    epilog=podman_epilog,
    options_metavar="[options]",
    passthrough="podman",
    invoke_without_command=True,
)
@click.pass_context
@click.option(
    "--additional-stores", type=str, help="Specify other storage locations"
)
@click.option(
    "--squash-dir",
    type=str,
    help="Specify alternate squash directory location",
)
@click.option("--log-level", type=str, default="fatal", hidden=True)
def podhpc(ctx, additional_stores, squash_dir, log_level):
    """Manage pods, containers and images ... on HPC!

    The podman-hpc utility is a wrapper script around the podman
    container engine. It provides additional subcommands for ease of
    use and configuration of podman in a multi-node, multi-user high
    performance computing environment.

    """
    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help())
        ctx.exit()

    # set up site configuration object
    try:
        conf = SiteConfig(squash_dir=squash_dir, log_level=log_level)
    except Exception as ex:
        sys.stderr.write(f"Error: {ex}... Exiting\n")
        sys.exit(1)

    if not os.path.exists(conf.squash_dir):
        ImageStore(conf.squash_dir, read_only=False).init_storage()
    conf.read_site_modules()
    conf.config_env(hpc=True)

    # add appropriate flags to call_podman based on invoked subcommand
    # defcmd = ctx.command.default_command_fn
    invcmd = ctx.command.get_command(ctx, ctx.invoked_subcommand)
    for k, v in conf.sitemods.get(ctx.invoked_subcommand, {}).items():
        if 'cli_arg' not in v:
            continue
        invcmd = click.option(
            f"--{v['cli_arg']}",
            is_flag=True,
            hidden=v.get("hidden", False),
            help=v.get("help"),
        )(invcmd)

    # save the site config to a context object so it can be passed to
    # subcommands
    ctx.obj = conf


# podman-hpc infohpc subcommand ############################################
@podhpc.command(options_metavar="[options]")
@pass_siteconf
def infohpc(siteconf):
    """Dump configuration information for podman_hpc."""
    siteconf.dump_config()
    sys.exit()


# podman-hpc migrate subcommand ############################################
@podhpc.command(options_metavar="[options]")
@pass_siteconf
@click.argument("image", type=str)
def migrate(siteconf, image):
    """Migrate an image to squashed."""
    mu = MigrateUtils(conf=siteconf)
    mu.migrate_image(image)
    sys.exit()


# podman-hpc rmsqi subcommand ##############################################
@podhpc.command(options_metavar="[options]")
@pass_siteconf
@click.argument("image", type=str)
def rmsqi(siteconf, image):
    """Removes a squashed image."""
    mu = MigrateUtils(conf=siteconf)
    mu.remove_image(image)

# podman-hpc images subcommand #############################################
@pass_siteconf
@click.pass_context
@click.argument("podman_args", nargs=-1, type=click.UNPROCESSED)
def images(ctx, siteconf, image, podman_args, **site_opts):
    """Displays images in both local and additionalimagestore."""
    cmd = [siteconf.podman_bin, "images"]
    cmd.extend(podman_args)
    cmd.extend(siteconf.get_cmd_extensions("images", site_opts))

# podman-hpc pull subcommand (modified) ####################################
@podhpc.command(
    context_settings=dict(
        ignore_unknown_options=True,
    ),
    options_metavar="[options]",
)
@pass_siteconf
@click.pass_context
@click.argument("podman_args", nargs=-1, type=click.UNPROCESSED)
@click.argument("image")
def pull(ctx, siteconf, image, podman_args, **site_opts):
    """Pulls an image to a local repository and makes a squashed copy."""
    cmd = [siteconf.podman_bin, "pull"]
    cmd.extend(podman_args)
    cmd.extend(siteconf.get_cmd_extensions("pull", site_opts))
    cmd.append(image)
    proc = Popen(cmd)
    proc.communicate()
    if proc.returncode == 0:
        sys.stdout.write(f"INFO: Migrating image to {siteconf.squash_dir}\n")
        mu = MigrateUtils(conf=siteconf)
        mu.migrate_image(image)
    else:
        sys.stderr.write("Pull failed.\n")
        sys.exit(proc.returncode)

# podman-hpc shared-run subcommand #########################################
@podhpc.command(
    context_settings=dict(
        ignore_unknown_options=True,
    ),
    options_metavar="[options]",
)
@pass_siteconf
@click.argument(
    "run-args",
    nargs=-1,
    type=click.UNPROCESSED,
    metavar="IMAGE [COMMAND [ARG...]]",
)
def shared_run(conf, run_args, **site_opts):
    """Launch a single container and exec many threads in it

    This is the recommended way to launch a container from a parallel launcher
    such as Slurm `srun` or `mpirun`. One container will be started (per node),
    and all process tasks will then be launched into that container via `exec`
    subcommand.

    When all `exec` processes have concluded, the container will close and
    remove itself.

    For valid flag options, see help pages for `run` and `exec` subcommands:

    \b
      podman-hpc run --help
      podman-hpc exec --help
    """
    # click.echo(f"Launching a shared-run with args: {sys.argv}")
    _shared_run(conf, run_args, **site_opts)

def _shared_run(conf, run_args, **site_opts):
    """
    Internal function for the shared_run.  This is so we can
    also call it when the user does run but enabled a module
    that has shared_run set to True. 
    """

    localid = os.environ.get(conf.localid_var)
    ntasks_raw = os.environ.get(conf.tasks_per_node_var, "1")
    ntasks = int(re.search(conf.ntasks_pattern, ntasks_raw)[0])
    container_name = f"uid-{os.getuid()}-pid-{os.getppid()}"
    sock_name = f"/tmp/uid-{os.getuid()}-pid-{os.getppid()}"

    # construct run and exec commands from user options
    # We need to filter out any run args in the run_args
    cmd = [conf.podman_bin, "run", "--help"]
    valid_params = cpt.filterValidOptions(list(run_args), cmd)
    # Find the first occurence not in the valid list
    idx = 0
    for idx, item in enumerate(run_args):
        if item in valid_params:
            continue
        break
    image = run_args[idx]
    container_cmd = run_args[idx+1:]
    # TODO: maybe do some validation on the iamge and container_cmd

    options = sys.argv[
        sys.argv.index("shared-run") + 1: sys.argv.index(image)
    ]

    run_cmd = [conf.podman_bin, "run", "--rm", "-d", "--name", container_name]
    run_cmd.extend(
        cpt.filterValidOptions(options, [conf.podman_bin, "run", "--help"])
    )
    run_cmd.extend(conf.get_cmd_extensions("run", site_opts))
    run_cmd.append(image)
    run_cmd.extend(conf.shared_run_command)

    exec_cmd = [
        conf.podman_bin,
        "exec",
    ]
    exec_cmd.extend(conf.get_cmd_extensions("exec", site_opts))
    exec_cmd.extend(pmi_fd())
    exec_cmd.extend(conf.shared_run_exec_args)
    exec_cmd.extend(
        cpt.filterValidOptions(options, [conf.podman_bin, "exec", "--help"])
    )
    exec_cmd.extend([container_name] + list(container_cmd))
    # click.echo(f"run_cmd is: {run_cmd}")
    # click.echo(f"exec_cmd is: {exec_cmd}")

    # Start monitor and run threads
    monitor_thread = None
    run_thread = None
    proc = None
    if (localid is None or int(localid) == 0):
        monitor_thread = Process(target=monitor, args=(sock_name, ntasks,
                                                       container_name, conf))
        monitor_thread.start()
        run_thread = Process(target=shared_run_exec, args=(run_cmd, conf.env))
        run_thread.start()

    try:
        # wait for container to exist
        comm = ["container", "exists", container_name]
        start_time = time.time()
        while podman_devnull(comm, conf) != 0:
            time.sleep(conf.wait_poll_interval)
            if time.time() - start_time > conf.wait_timeout:
                msg = "Timeout waiting for shared-run start"
                raise OSError(msg)
            if run_thread and run_thread.exitcode:
                raise OSError("Failed to start container")
        comm = ["wait", "--condition", "running", container_name]
        podman_devnull(comm, conf)
        fds = [0, 1, 2]
        if 'PMI_FD' in os.environ:
            fds.append(int(os.environ['PMI_FD']))
            conf.env["PMI_FD"] = os.environ["PMI_FD"]
        proc = Popen(exec_cmd, env=conf.env, pass_fds=fds)
        proc.communicate()
        send_complete(sock_name, localid)
        # Close out threads
        if monitor_thread:
            monitor_thread.join()
        if run_thread:
            run_thread.join()
    except Exception as ex:
        sys.stderr.write(str(ex))
        if monitor_thread:
            sys.stderr.write("Killing monitor thread")
            monitor_thread.kill()
        if run_thread:
            run_thread.kill()
        if os.path.exists(sock_name):
            os.remove(sock_name)
    finally:
        exit_code = 1
        if proc:
            exit_code = proc.returncode
        sys.exit(exit_code)


# podman-hpc call_podman subcommand (default, hidden, passthrough) #########
@podhpc.default_command(
    context_settings=dict(ignore_unknown_options=True, help_option_names=[]),
    hidden=True,
)
@pass_siteconf
@click.pass_context
@click.option("--help", is_flag=True, hidden=True)
@click.argument("podman_args", nargs=-1, type=click.UNPROCESSED)
def call_podman(ctx, siteconf, help, podman_args, **site_opts):
    cmd = [siteconf.podman_bin, ctx.info_name]
    cmd.extend(siteconf.get_cmd_extensions(ctx.info_name, site_opts))
    cmd.extend(pmi_fd())
    cmd.extend(podman_args)

    # if the help flag is called, we pass podman's STDOUT stream through a pipe
    # to a stream editor, to inject additional help page info
    if help:
        app_name = ctx.find_root().command_path
        passthrough_name = os.path.basename(cmd[0])

        formatter = ctx.make_formatter()
        ctx.command.format_options(ctx, formatter)
        app_options = ""
        if formatter.getvalue():
            app_options = (
                f"{app_name.capitalize()} {formatter.getvalue()}\n")

        cmd = [siteconf.podman_bin, ctx.info_name, "--help"]
        proc = Popen(cmd, env=siteconf.env, stdout=PIPE)
        out, _ = proc.communicate()
        option_line = f"{app_options}{passthrough_name.capitalize()}"
        option_line += " options follow:"
        newout = ""
        for line in out.decode().split("\n"):
            newline = line.replace(passthrough_name, app_name)
            newline = newline.replace("Options:", option_line)
            newout += f"{newline}\n"

        sys.stdout.write(newout)
    else:
        if siteconf.shared_run:
            for idx, arg in enumerate(sys.argv):
                if arg == "run":
                    sys.argv[idx] = "shared-run"
            _shared_run(siteconf, podman_args, **site_opts)
        else:
            if 'PMI_FD' in os.environ:
                siteconf.env["PMI_FD"] = os.environ["PMI_FD"]
            os.execve(cmd[0], cmd, siteconf.env)


def shared_run_exec(run_cmd, env):
    proc = Popen(run_cmd, stdout=PIPE, stderr=PIPE, env=env)
    out, err = proc.communicate()
    if proc.returncode != 0:
        sys.stderr.write(err.decode())


def monitor(sockfile, ntasks, container_name, conf):
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        os.remove(sockfile)
    except OSError:
        pass
    s.bind(sockfile)
    ct = 0
    while True:
        s.listen()
        conn, addr = s.accept()
        ct += 1
        if ct == ntasks:
            break
    conn.close()
    os.remove(sockfile)
    # cleanup
    podman_devnull(["kill", container_name], conf)
    podman_devnull(["rm", container_name], conf)


def send_complete(sockfile, lid):
    try:
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        s.connect(sockfile)
        s.send(bytes(lid, 'utf-8'))
        s.close()
    except Exception as ex:
        sys.stderr.write(f"send_complete failed for {lid}\n{ex}\n")


def main():
    podhpc(prog_name="podman-hpc")


if __name__ == "__main__":
    main()



#!/usr/bin/python3
import ctypes
import ctypes.util
import os
import sys
import errno
import subprocess
import json
import yaml
import shutil
import re
from glob import glob, iglob

_MOD_ENV = "PODMANHPC_MODULES_DIR"

_libc = ctypes.CDLL(ctypes.util.find_library("c"), use_errno=True)
_libc.mount.argtypes = (
    ctypes.c_char_p,
    ctypes.c_char_p,
    ctypes.c_char_p,
    ctypes.c_ulong,
    ctypes.c_char_p,
)
logger = None


def log(msg):
    if logger:
        logger.write(f"{msg}\n")


def setns(pid, ns):
    """
    Attach to a namespace using the pid
    ns = ["mnt", "pid", "net", ...]
    """
    pth = f"/proc/{pid}/ns/{ns}"
    tgt = open(pth, "r")
    if _libc.setns(tgt.fileno(), 0) == -1:
        e = ctypes.get_errno()
        raise OSError(e, errno.errorcode[e])


def bind_mount(src, tgt):
    """
    bind mount a file into a namespace.
    """
    # Create mount point
    if os.path.isdir(src) and not os.path.exists(tgt):
        os.makedirs(tgt)
    elif not os.path.exists(tgt):
        open(tgt, "w").close()
    subprocess.check_output(["mount", "--rbind", src, tgt])


def copy(src, tgt, symlinks=True):
    if os.path.isdir(src):
        shutil.copytree(
            src,
            tgt,
            symlinks=symlinks,
            copy_function=shutil.copyfile,
            dirs_exist_ok=True,
        )
    else:
        shutil.copyfile(src, tgt, follow_symlinks=(not symlinks))


def ldconfig():
    if not os.path.exists("/sbin/ldconfig"):
        return
    ret = "unknown"
    try:
        ret = subprocess.check_output(["/sbin/ldconfig"])
    except subprocess.CalledProcessError:
        log(f"ldconfig failed: {ret}")


def resolve_src_and_dest(rule, root_path, modulesd=os.path.abspath("")):
    # extract source and destination patterns from the rule
    rs, rd = (rule + ":").split(":")[:2]

    # expand vars
    rs = os.path.expanduser(os.path.expandvars(rs))
    rd = os.path.expanduser(os.path.expandvars(rd))

    # ensure source pattern is absolute path
    if not os.path.isabs(rs):
        rs = os.path.join(modulesd, rs)
    rs = os.path.abspath(os.path.expanduser(os.path.expandvars(rs)))

    # error checks
    if not os.path.isabs(rd):
        log(
            f"Error: Destination in pattern must be an absolute path.\n\tdestination: {rd}"
        )
        return {}
    if "*" in rd:
        if rs.count("*") != 1:
            log(
                "Error: Using glob '*' in destination requires exactly one glob '*' in source."
                f"\n\tsource: {rs}\n\tdestination: {rd}"
            )
            return {}
        else:
            # this is used to capture the glob expansion later
            globstrip = "^{}|{}$".format(*rs.split("*"))

    # determine how to construct absolute path to destination object
    # if no destination rule is given, use the path from source rule
    if rd == "":
        dest = lambda src: os.path.join(root_path, src[1:])
    # if the destination rule reuses the glob pattern, trailing path separator
    # determines if it interpreted as a directory or obj name
    elif "*" in rd and rd[-1] == os.path.sep:
        dest = lambda src: os.path.join(
            root_path,
            rd.replace("*", re.sub(globstrip, "", src))[1:],
            os.path.basename(src),
        )
    elif "*" in rd and rd[-1] != os.path.sep:
        dest = lambda src: os.path.join(
            root_path, rd.replace("*", re.sub(globstrip, "", src))[1:]
        )
    # if we glob multiple sources, interpret destination as a directory, and append shortest unique path
    elif 1 < len(glob(rs)):
        common_prefix = os.path.commonpath(glob(rs))
        dest = lambda src: os.path.join(
            root_path, rd[1:], os.path.relpath(src, common_prefix)
        )
    # otherwise, trailing path separator determines whether destination is interpreted as a directory
    elif rd[-1] == os.path.sep:
        dest = lambda src: os.path.join(
            root_path, rd[1:], os.path.basename(src)
        )
    else:
        dest = lambda src: os.path.join(root_path, rd[1:])

    return {src: os.path.normpath(dest(src)) for src in iglob(rs)}


def do_plugin(rp, mod, modulesd):
    """
    set up to do copies and bind mounts
    handle wildcards appropriately
    """
    log(f"Module: {mod}")

    actions = {"copy": copy, "bind": bind_mount}

    for a in actions:
        for rule in mod.get(a) or []:
            log(f"\t{rule}")
            for src, tgt in resolve_src_and_dest(rule, rp, modulesd).items():
                os.makedirs(os.path.dirname(tgt), exist_ok=True)
                log(f"\t\t{a}: {src} to {tgt}")
                actions[a](src, tgt)


def read_confs(mdir):
    confs = {}
    for d in glob(f"{mdir}/*.yaml"):
        conf = yaml.load(open(d), Loader=yaml.FullLoader)
        confs[conf["name"]] = conf
    return confs


def main():
    global logger

    inp = json.load(sys.stdin)
    pid = inp["pid"]
    cf = json.load(open("config.json"))
    cf_env = {}
    # initialize with any values set in the hook env configuration
    cf_env.update(os.environ)
    for e in cf["process"]["env"]:
        k, v = e.split("=", maxsplit=1)
        cf_env[k] = v

    plug_conf_fn = cf_env.get(
        _MOD_ENV, f"{sys.prefix}/etc/podman_hpc/modules.d"
    )
    plug_conf = read_confs(plug_conf_fn)

    lf = cf_env.get("LOG_PLUGIN")
    if lf:
        logger = open(lf, "w")
    log("input")
    log(json.dumps(inp, indent=2))
    log("config.json")
    log(json.dumps(cf, indent=2))
    rp = cf["root"]["path"]

    setns(pid, "mnt")
    os.chroot("/")
    log(json.dumps(plug_conf, indent=2))
    for m in plug_conf:
        if plug_conf[m]["env"] in cf_env:
            log(f"Loading {m}")
            do_plugin(rp, plug_conf[m], plug_conf_fn)
    ret = os.chroot(rp)
    log(f"chroot return: {ret}")

    ldconfig()


if __name__ == "__main__":
    main()



from podman_hpc.migrate2scratch import MigrateUtils
import os
import json
import pytest
from tempfile import TemporaryDirectory


def get_count(fn, img):
    images = json.load(open(fn))
    ct = 0
    for i in images:
        for n in i['names']:
            if n == img:
                ct += 1
    return ct


class mockproc():
    returncode = 0

    def __init__(self, rcode=None):
        if rcode:
            self.returncode = rcode

    def communicate(self):
        return b"blah", b"blah"


@pytest.fixture
def src():
    tdir = os.path.dirname(__file__)
    return os.path.join(tdir, "storage")


@pytest.fixture
def dst():
    tempd = TemporaryDirectory(dir="/tmp")
    return tempd.name


def test_init_storage(src):
    with TemporaryDirectory() as dst:
        mu = MigrateUtils(src=src, dst=dst)
        mu._lazy_init()
        mu.dst.init_storage()
        idir = os.path.join(dst, "overlay-images")
        assert os.path.exists(idir)


def test_bad_image_name(src):
    with TemporaryDirectory() as dst:
        mu = MigrateUtils(src=src, dst=dst)
        mu.migrate_image("balpine")


def test_migrate_remove(src, tmp_path, mocker):
    img = "docker.io/library/alpine:latest"
    hash = "9c6f0724472873bb50a2ae67a9e7adcb57673a183cea8b06eb778dca859181b5"
    tdir = os.path.dirname(__file__)
    bimg = json.load(open(os.path.join(tdir, "bogus_image.json")))

    # Mock Popen
    popen = mocker.patch("podman_hpc.migrate2scratch.Popen")
    mu = MigrateUtils(src=src, dst=tmp_path)
    mu._lazy_init()
    mu.dst.init_storage()
    imgf = os.path.join(tmp_path, "overlay-images/images.json")
    with open(imgf, "w") as f:
        json.dump([bimg], f)
    mu.dst.refresh()

    # Mock squash failing
    popen.return_value = mockproc(rcode=1)
    resp = mu.migrate_image(img)
    assert resp is False

    # Now a successful one
    popen.return_value = mockproc()
    resp = mu.migrate_image(img)
    assert resp
    assert get_count(mu.dst.images_json, img) == 1
    popen.assert_called()

    # Remigrate to test check logic
    resp = mu.migrate_image(img)
    assert resp
    assert get_count(mu.dst.images_json, img) == 1

    # Remigrate with hash
    resp = mu.migrate_image(hash)
    assert resp
    assert get_count(mu.dst.images_json, img) == 1

    # Test removing the image
    resp = mu.remove_image(img)
    assert resp
    assert get_count(mu.dst.images_json, img) == 0



#!/usr/bin/env python
import subprocess
from setuptools import setup
from setuptools.command.build_py import build_py


class BuildMakeCommand (build_py):
    def run(self):
        super(build_py, self).run()
        subprocess.run(['make', 'setuptools-build_py'], check=True)


if __name__ == "__main__":
    setup(cmdclass={'build_py': BuildMakeCommand})



#!/usr/bin/python
import os
import sys
import json
from shutil import copytree, copy, which
from subprocess import Popen, PIPE
import logging

DEBUG = os.environ.get("DEBUG_M2SQ", False)


def merge_recs(recs_list, key):
    """
    Generic method to merge a list of records
    """
    res = []
    done = {}
    for recs in recs_list:
        for r in recs:
            id = r[key]
            if id not in done:
                res.append(r)
                done[id] = r
    return res


class ImageStore:
    """
    Class to provide some basic functions for interacting with
    an image store.
    """

    images = []
    layers = []

    def __init__(self, base, read_only=True):
        """
        Inputs:
        base: base directory path of the image store
        """
        self.base = base
        self.images_dir = os.path.join(base, "overlay-images")
        self.images_json = os.path.join(self.images_dir, "images.json")
        self.layers_dir = os.path.join(base, "overlay-layers")
        self.layers_json = os.path.join(self.layers_dir, "layers.json")
        self.overlay_dir = os.path.join(base, "overlay")
        self.read_only = read_only
        if os.path.exists(self.images_json):
            self.images = json.load(open(self.images_json))
        if os.path.exists(self.layers_json):
            self.layers = json.load(open(self.layers_json))

    def refresh(self):
        """
        Currently just used in testing.
        """
        if os.path.exists(self.images_dir):
            self.images = json.load(open(self.images_json))
            self.layers = json.load(open(self.layers_json))

    def get_img_info(self, img_name):
        """
        Finds the image info by name.  This tries different variant
        including adding repositories and latest tag.

        Inputs:
        img_name: name to lookup. Can be a short form.
        """
        # Try exact match
        for img in self.images:
            if img["id"].startswith(img_name):
                logging.debug("Found by ID")
                return img, img["id"]
        if ":" not in img_name:
            img_name = f"{img_name}:latest"
        prefs = ["", "docker.io/", "docker.io/library/", "localhost/"]
        for pref in prefs:
            long_name = f"{pref}{img_name}"
            for img in self.images:
                for n in img.get("names", []):
                    if long_name == n:
                        return img, long_name
        return None, None

    def get_manifest(self, imgid):
        """
        Retruns the contents of the manifest for the given image ID

        Inputs:
        imgid: image id
        """
        mf = os.path.join(self.images_dir, imgid, "manifest")
        return json.load(open(mf))

    def init_storage(self):
        """
        Initializes a directory as an image store.  This creates
        the minimum directories and JSON files.
        """
        if self.read_only:
            raise ValueError("Cannot init read-only stroage")

        if not os.path.exists(self.base):
            os.mkdir(self.base)
        for ext in ["", "/l", "-images", "-layers"]:
            pth = os.path.join(self.base, f"overlay{ext}")
            if not os.path.exists(pth):
                os.mkdir(pth)
        for typ in ["images", "layers"]:
            pth = f"{self.base}/overlay-{typ}/{typ}.lock"
            if not os.path.exists(pth):
                with open(pth, "w") as f:
                    f.write("")
            pth = f"{self.base}/overlay-{typ}/{typ}.json"
            if not os.path.exists(pth):
                with open(pth, "w") as f:
                    f.write("[]")

    def chk_image(self, id):
        """
        Checks if an ID is present in a image store. Returns True/False

        Inputs:
        id: Image ID
        """
        for img in self.images:
            if img["id"] == id:
                return True
        return False

    def del_rec(self, otype, id, key="id"):
        """
        Deletes a record from a JSON file

        Inputs:
        otype: object type (images/layers)
        id: object id
        key: key name for the ID
        """
        if self.read_only:
            raise ValueError("Cannot init read-only stroage")

        fn = os.path.join(self.base, f"overlay-{otype}", f"{otype}.json")
        data = json.load(open(fn))
        changed = False
        out = []
        for rec in data:
            if rec[key] == id:
                changed = True
                continue
            out.append(rec)
        if changed:
            json.dump(out, open(fn, "w"))
            logging.debug(f"Updated {fn}")

    def drop_tag(self, image, id):
        """
        Removes an image tag from an image by its ID.
        This leaves the repo but just drops the tag which
        is only allowed to be set for one image.

        Inputs:
        image: image tag name
        id: image ID
        """
        if self.read_only:
            raise ValueError("Cannot init read-only stroage")

        data = self.images
        for img in data:
            if img["id"] == id:
                nnames = []
                for name in img["names"]:
                    if name == image:
                        name = ":".join(image.split(":")[:-1])
                    nnames.append(name)
                img["names"] = nnames
        json.dump(data, open(self.images_json, "w"))
        self.images = data

    def add_recs(self, otype, recs):
        """
        Adds records to the JSON store.

        Inputs:
        otype: Object type (images/layers)
        recs: list of records to add.

        Note: No validation is done on the records other than checking
              for duplicate IDs.
        """
        if self.read_only:
            raise ValueError("Cannot init read-only stroage")

        fn = os.path.join(self.base, f"overlay-{otype}", f"{otype}.json")
        data = json.load(open(fn))
        by_id = {}
        for row in data:
            by_id[row["id"]] = row

        changed = False
        for rec in recs:
            if rec["id"] not in by_id:
                data.append(rec)
                changed = True
        if changed:
            json.dump(data, open(fn, "w"))
            logging.debug(f"Updated {fn}")
            self.refresh()

    def get_squash_filename(self, link):
        return os.path.join(self.overlay_dir, "l", f"{link}.squash")

    def read_link_file(self, img_id):
        """
        Read the overlay link file
        """
        lf = os.path.join(self.overlay_dir, img_id, "link")
        return open(lf).read()


class MigrateUtils:
    """
    Utility to migrate/copy images from one image store to another.
    """

    src = None
    dst = None
    images = None
    podman_bin = "podman"
    mksq_bin = "mksquashfs.static"
    mksq_options = ["-comp", "lz4", "-xattrs-exclude", "security.capability"]
    exclude_list = ["/sqout", "/mksq", "/proc", "/sys", "/dev"]
    _mksq_inside = "/mksq"

    def __init__(self, src=None, dst=None, conf=None):
        """
        Inputs:
        src: base directory of source image store
        dst: base directory of destination image store
        conf: a podman_hpc config object

        If src isn't provided, then default to user's default store.

        If dst isn't provided, then default to the values of the
        SQUASH_DIR environment variable.
        """
        self.src_dir = src
        self.dst_dir = dst
        self._lazy_init_called = False
        if conf:
            self.podman_bin = conf.podman_bin
            self.mksq_bin = conf.mksquashfs_bin
            if not self.src_dir:
                self.src_dir = conf.graph_root
            if not self.dst_dir:
                self.dst_dir = conf.squash_dir

    def _lazy_init(self):
        if not self._lazy_init_called:
            self.src_dir = self.src_dir or self._get_paths()
            self.src = ImageStore(self.src_dir)
            self.dst_dir = self.dst_dir or os.environ["SQUASH_DIR"]
            self.dst = ImageStore(self.dst_dir, read_only=False)
            self._lazy_init_called = True

    @staticmethod
    def _get_paths():
        """
        Helper function to lookup the default image store.
        """
        home = os.environ["HOME"]
        cf = f"{home}/.config/containers/storage.conf"
        with open(cf) as f:
            for line in f:
                if "#" in line:
                    continue
                if "graphroot" in line:
                    val = line.rstrip().split("=")[1]
                    p = val.replace(" ", "").replace('"', "")
        return p

    def _get_img_layers(self, store, imgid):
        """
        This finds all the required layers for an image
        including layers coming from dependent images.

        Inputs:
        imgid: Image ID
        """

        def _add_parent(layer, layers, by_id, layer_ids):
            """
            Recrusive function to walk up parent graph.

            Inputs:
            layer: layer to walk
            layers: list of layers that are being accumulated.
            by_id: dictionary of layers by ID
            layer_ids: accumulated dictionary of layers by ID
            """
            if "parent" in layer and layer["parent"] not in layer_ids:
                parent = by_id[layer["parent"]]
                layers.append(parent)
                layer_ids.add(parent["id"])
                _add_parent(parent, layers, by_id, layer_ids)

        by_digest = {}
        by_id = {}
        all_layers = merge_recs([self.src.layers, self.dst.layers], "id")
        for layer in all_layers:
            if "compressed-diff-digest" in layer:
                by_digest[layer["compressed-diff-digest"]] = layer
            if "diff-digest" in layer:
                by_digest[layer["diff-digest"]] = layer
            by_id[layer["id"]] = layer
        md = store.get_manifest(imgid)
        layers = []
        layer_ids = set()
        for layer in md["layers"]:
            ld = by_digest[layer["digest"]]
            layer_ids.add(ld["id"])
            _add_parent(ld, layers, by_id, layer_ids)
            layers.append(ld)

        return layers

    def _copy_image_info(self, img_id):
        srcd = os.path.join(self.src.images_dir, img_id)
        dstd = os.path.join(self.dst.images_dir, img_id)
        # Copy image directory
        if not os.path.exists(dstd):
            copytree(srcd, dstd)

    def _copy_required_layers(self, req_layers):
        for layer in req_layers:
            layer_id = layer["id"]
            fn = f"{layer_id}.tar-split.gz"
            srcd = os.path.join(self.src.layers_dir, fn)
            dstd = os.path.join(self.dst.layers_dir, fn)
            if not os.path.exists(dstd):
                logging.debug(f"Copy {srcd} to {dstd}")
                copy(srcd, dstd)
        self.dst.add_recs("layers", req_layers)

    def _copy_overlay(self, img_id, layers):
        for layer in layers:
            id = layer["id"]
            sbpath = os.path.join(self.src.overlay_dir, id)
            dbpath = os.path.join(self.dst.overlay_dir, id)
            if os.path.exists(sbpath) and not os.path.exists(dbpath):
                os.mkdir(dbpath)
            for p in ["empty", "merged", "work", "diff"]:
                spath = os.path.join(sbpath, p)
                dpath = os.path.join(dbpath, p)
                if os.path.exists(spath) and not os.path.exists(dpath):
                    os.mkdir(dpath)
            # the link
            src = os.path.join(sbpath, "link")
            dst = os.path.join(dbpath, "link")
            if not os.path.exists(dst):
                logging.debug(f"Copy {src} to{dst}")
                copy(src, dst)

            # Create symlink file
            link = self.dst.read_link_file(id)
            lname = os.path.join(self.dst.overlay_dir, "l", link)
            tgt = os.path.join("..", id, "diff")
            if not os.path.exists(lname):
                os.symlink(tgt, lname)
            # Finally the squash file
            # Since there typically isn't a squash file, this is more
            # for future cases
            src = self.src.get_squash_filename(link)
            dst = self.dst.get_squash_filename(link)
            if os.path.exists(src) and not os.path.exists(dst):
                logging.debug(f"Copy {src} to {dst}")
                copy(src, dst)

    def _mksq(self, img_id, top_id):
        # Get the link name
        ln = self.dst.read_link_file(top_id)
        _mksqstatic = self.mksq_bin
        if not _mksqstatic.startswith("/"):
            _mksqstatic = which(_mksqstatic)
        tgt = self.dst.get_squash_filename(ln)
        if os.path.exists(tgt):
            logging.info("Squash file already generated")
            return True
        logging.info(f"Generating squash file {tgt}")
        # To make the squash file we will start up a container
        # with the tgt image and then run mksq in it.
        # This requires a statically linked mksquashfs
        com = [
            self.podman_bin, "run", "--rm",
            "--root", self.src.base,
            "-v", f"{_mksqstatic}:{self._mksq_inside}",
            "-v", f"{self.dst.base}/overlay/l/:/sqout",
            "--user", "0",
            "--entrypoint", self._mksq_inside,
            img_id,
            "/", f"/sqout/{ln}.squash",
        ]
        com.extend(self.mksq_options)
        # Exclude these
        for ex in self.exclude_list:
            com.extend(["-e", ex])
        proc = Popen(com, stdout=PIPE, stderr=PIPE, env=os.environ)
        out, err = proc.communicate()

        if proc.returncode != 0:
            logging.error("Squash Failed")
            logging.error(out.decode("utf-8"))
            logging.error(err.decode("utf-8"))
            return False

        logging.info("Created squash image")
        return True

    def migrate_image(self, image):
        self._lazy_init()
        logging.debug(f"Migrating {image}")
        self.dst.init_storage()
        self.src.refresh()
        self.dst.refresh()
        # Read in json data

        img_info, fullname = self.src.get_img_info(image)
        if not img_info:
            logging.error(f"Image {image} not found\n")
            return False

        img_id = img_info["id"]
        # Get the layers from the manifest
        rld = self._get_img_layers(self.src, img_id)

        # make sure the src squash file exist
        top_id = rld[-1]["id"]
        logging.debug(f"Reading link: {top_id}")

        if self.dst.chk_image(img_id):
            logging.info("Previously migrated")
            return True

        # Check if previously tagged image exist
        dimg = None
        if fullname:
            dimg, _ = self.dst.get_img_info(fullname)
        if dimg and dimg["id"] != img_info["id"]:
            logging.info("Replace previous version")
            self.dst.drop_tag(fullname, dimg["id"])

        # Copy image info
        self._copy_image_info(img_id)

        # Copy layers
        self._copy_required_layers(rld)

        # Overlay
        self._copy_overlay(img_id, rld)

        # Generate squash
        logging.debug(f"squashing {img_id}")
        resp = self._mksq(img_id, top_id)
        if not resp:
            return False

        # Add img to images.json
        # Save this for the end so things are all ready
        self.dst.add_recs("images", [img_info])
        return True

    def remove_image(self, image):
        self._lazy_init()
        logging.debug(f"Removing {image}")
        self.dst.refresh()
        img_info, _ = self.dst.get_img_info(image)
        if not img_info:
            logging.error(f"Image {image} not found\n")
            return False
        img_id = img_info["id"]
        # Get the layers from the manifest
        rld = self._get_img_layers(self.dst, img_id)

        # make sure the src squash file exist
        top_id = rld[-1]["id"]
        ln = self.dst.read_link_file(top_id)
        sqf = self.dst.get_squash_filename(ln)
        if os.path.exists(sqf):
            os.unlink(sqf)
        self.dst.del_rec("images", img_id)
        return True


def usage():
    """
    print usage info
    """
    print("Usage: m2scr [mig|rmi|init] <image name> [<dest>]")
    print("Set SQUASH_DIR to define the default destination")


if __name__ == "__main__":  # pragma: no cover
    mu = MigrateUtils()
    dst = os.environ.get("SQUASH_DIR")
    if len(sys.argv) < 2:
        usage()
    elif sys.argv[1] == "rmi":
        sys.argv.pop(0)
        image = sys.argv[1]
        if not dst:
            dst = sys.argv[2]
        mu.remove_image(image)
        sys.exit()
    elif sys.argv[1] == "init":
        if not dst:
            dst = sys.argv[2]
        mu.dst.init_storage(dst)
    elif sys.argv[1].startswith("mig"):
        image = sys.argv[2]
        if not dst:
            dst = sys.argv[3]
        mu.migrate_image(image, dst)
    else:
        usage()
